( * 3iBeats.fs - AE )
( Firmware: 
    Requires optionally:
        3iForth.avr  or 
        if using matrix keypad then 3iKeypad.avr   or 
        if using integrated menu buttons w/ OLED then 3iOled.avr if Adafruit driver or 3iI2C.avr if my lightweight driver
        if using shift registers ... 
)
(   It does not require 3iBeats.avr which is for direct generation of fixed beat into SFX 76477 board )

( Applications:
    Metronome with variable pace setting
)


\ HARDWARE:
\   speaker on A0 / pin14 of Nano
\   OLED SSD1306 in pins A4/A5 of Nano
\   5 buttons (menu) on pins D2 to D6
\   timing potentiometer on pins 15/A1 and 21/A7
14 constant buzzer  \ connect speaker to A0 pin 14

\ Potentiometer pins are 15/A1 (out) -> (via resistor) to A7 (analog in) -> pot (A10K pot - audio A taper pot has better performance in low range)
15 constant potout \ pin 15 of Nano
21 constant potin  \ A7 of Nano


\ TO PC SCREEN  parses bits of a given hex word
22 constant thickbar
14 constant musicnote
45 constant rest
: .star ( n -- ) 0> if '*' else bl then emit ;  \ star and blank
: .stars (  n  --  )  \  print stars/blanks - loop through bits from MSB mask and print  \ used in 3iFont.fs
    T8 init-mask begin 2dup and .star 1 rshift dup 0= until 2drop ;
: .note-bit ( n -- ) 0> if musicnote  else rest then emit ;  \ note and rest
: .notes-word (  n T4/T8 --  )  \  print notes/bars loop through bits from MSB mask and print   \ used in 3iBeat.fs and 3iMenu.fs
    init-mask begin 2dup and .note-bit 1 rshift dup 0= until 2drop 2 spaces ;
variable notecount
: reset-notecount 0 notecount ! ;
: notecount++ ( -- n++ ) notecount @ 1 + dup notecount ! ;

: [visualize-beats]  ( <beats-hex> N -- )   \ process from top-of-stack
    0 do 
         T4 .notes-word space notecount++ 
         8 = if cr then   \ each line has a 2-bar phrase
    loop ;
: visualize-beats ( n-elements N --  n-elements N ) \ displays notes-word sequence   e.g. 7 7 6 6     4 visualize-beats   \ used in beat sequencer, process left-to-right (bottom of stack)
    dup 0 > if 
       >r 
       cr reset-notecount 
       r@ ndup nrev [visualize-beats]  
       r> 
    else drop 
    then ;  
: nhex ( n-elements N --  n-elements ) \ displays stack in hex form  e.g. 7 7 6 6     4 visualize-beats   \ used in beat sequencer
    cr reset-notecount dup 0 > if
       dup 0 do dup i - pick hex. space notecount++ 8 = if cr then loop drop   \ 2-bar phrases on separate lines
    else drop 
    then ;  


\ ============ APPLICATION: ADJUST PACE WITH POTENTIOMETER ==================
variable beatdur
: beatdur! ( dur -- ) beatdur ! ;
: beatdur@ ( -- dur ) beatdur @ ;  \ TODO: have this be read by input POTENTIOMETER for rate control

: init-beats
  buzzer out
  100 beatdur!  \ default - 16th note at 100 ms
  potout out
  potout lo
; 

: readpot ( -- )  \ sets the 16th note beatdur from 0 to 200 ms  ( divider = 5 )
    potout hi potin pinreadLevel 2 / dup beatdur@ = if drop else dup . beatdur! then ;  
: setpot ( -- ) begin readpot key? until key drop ;

: // ( -- ) beatdur@ ms ;  \ Timings from Forth, 100ms = 16th note, 200ms = 8th note, 400ms = quarter note, and 800ms = half note, 1600ms = whole note

\ ============= APPLICATION: ADD TONAL CONTROL TO PERCUSSION BEATS              EXPERIMENTAL - 2021-05-31---Mon---23:33
\ Design: 
\ Beats and Rests are 2-bit music
\ Tonality gives greater color depth to music.
\ 2 types of drums gives 3-bit music
\ Tabla which depending on how it is played can have X different sounds gives X+1 bit music
\ We will provide either 8-bit music (7 octal tones) or 16-bit music (7 hex tones)
\ These tones can then be easily mapped to keys for a drum pad

\ Design:
\ Sounding tones.  We need to use the long-form below because the simple abbreviation for PCB Piano d_ note incorporated delay 
\ Beat synth needs minimum delay to throw the semiquavers and adjust tempo
\ Note assignments are from 3iSound-4x4.fs using c_ etc. to get the note index.  Usage:  e.g. d_ beat-tone
: beat-tone ( d_ -- ) freq lohi 20 lohi buzzer [tone] ;  \ 20 ms duration

\ Design: 
\ 1 Beat = 1 crotchet or 4 semiquavers, or 1 hex digit (4-bits, nibble)
\ 1 Bar = 4 crotchets or 16 semiquavers, and 32-bits (4 hex digits)
\ Percussion Rhythms are typically 2-bar or 4-bar phrases, so max 64 independent bits/tones for a 4-bar phrase
\ Phrases = 2 bars or 4 bars, Hook = 8 bars, Verse = 16 bars
create percustones 64 cells allot    
variable npercustones  \ number of tones
variable tonecount     \ pointer to tones
: percustones! ( n -- ) npercustones ! ;
: reset-tonecount 0 tonecount ! ;
: next-tone  tonecount @  1 +   npercustones @ mod   tonecount !  ;
: [play-tone] ( index -- ) 
    cells percustones + @ beat-tone ;
: play-next-tone ( -- )   \  plays percussion tones and cycles to next tone of percustones array - 16th note
    tonecount @  [play-tone] next-tone ;
: tones-demo ( n -- ) 0 do i [play-tone] // loop ;
: load-tones ( 1 2 3 4 n -- n )  \ provide notes as inputs
    >r r@ percustones! r@ nrev percustones swap 0 do !+ loop drop r> tones-demo ;
\ cyclic
: tones-fixed   ( -- <tones> n ) d_    d_  d_ d_ 4 ;   \ single tone beat
: tones-rasra   ( -- <tones> n ) d___  d_  d_ d_ 4 ;   \ for rasra3 dum-bateta
: tones-dchord  ( -- <tones> n ) a_    f_  d_ a_ 4 ;

\ TODO: experiment with arpeggios/chords from switching betweeen tones...

\ fixed position
: copy-beats ( tone n -- ) 0 do dup loop drop ;
: tones-uptown 
    d_ 26 copy-beats d' 6 copy-beats  
    d_ 26 copy-beats d' d_ d' 4 copy-beats  64
;

\ TODO: build keyboard for finger/drum kit pads allowing tapping the desired beat with tonality

: init-percustones
    100 beatdur!
    tones-fixed load-tones reset-tonecount
    ;

\ Design: percustones provides different toned percussion to be played in a loop (pulsetone)
\  Two types of rotation: cyclic, where tones are cycled regardless of how many rests are between, and 
\                         mapped, where tones are mapped to sections of a beat, e.g. the 1, or the 3 etc.
\                         mapped is more reflective of percussion, cyclic is reflective of chiptune music

\ ========== INSTRUMENTS
\ metronome as tempo selector has to time for every 4th 16th note, i.e. it sounds the beats of a bar (quarter notes)
: metronome  begin readpot play-next-tone // // // key? until key drop ;  \ metronome with pace control
: Nmetronome ( dur -- ) >r begin r@ ms ( buzzer pulse-hi ) play-next-tone key? until key drop r> drop ; 
\ TODO: Add precise beats per minute or beats per second keyed in interface...

\ Design:
\ Nibble or Byte entry? Nibble is better as it matches better the beat structure of music, 4 semiquavers to a crotchet (1 nibble)
\ MSB or LSB entry?  MSB is better as it matches the left to right notation that one would normally enter beats and expect to execute

\  \ BYTE ENTRY
\  \ beats: 237, 118, 17, 13, 213, 253, 0
\  \ seqs: 0 213 13 17    and   1 253 13 17 0 213 13 17
\  : beat-byte-lsb  ( b -- ) \ parse byte to beat - LSB first for shift in register  ex: 237 beat-byte-lsb
\      8 0 do dup 1 and if ( buzzer pulse-hi ) play-next-tone then // 1 rshift loop drop ;  
\  : Nbeat-lsb ( N-bytes N -- ) 0 do beat-byte-lsb loop ;  \  237 118  2 Nbeat
\  : Nseq-lsb ( N-bytes N -- N-bytes ) >r begin r@ ndup Nbeat-lsb readpot key? until key drop r> drop ;  \ PLAYS FROM TOS ie r2l
\  : demo-uptown-funk 100 beatdur! 1 253 13 17 0 213 13 17   8 Nseq-lsb ;  \ 16 notes = 4 bars uptown funky bunch: https://www.youtube.com/watch?v=M1F0lBnsnkE
\  
\  : beat-byte-msb  ( b -- )  \ parse byte to beat - MSB first for keyboard entry
\      9 1 do dup 1  8 i - lshift  and  8 i - rshift if ( buzzer pulse-hi ) play-next-tone then // loop drop ;
\  : Nbeat-msb ( N-bytes N -- N-bytes ) \ play sequence once
\      dup 0 do dup i - pick beat-byte-msb loop drop ;  \  156 170  2 Nbeat-msb
\  : Nseq-msb ( N-bytes N -- N-bytes ) \ loop sequence
\      >r r@ visualize-beats drop begin r@ Nbeat-msb ( readpot ) key? until key drop r> drop ;
\  

\ NIBBLE ENTRY - great for keypad ---- requires setting 3iForth.fs to T = $F  TODO: CREATE FUNCTION FOR SUCH CHANGE
\    then re-define visualize-beats for notes display, and the following 3
: beat-nibble-msb  ( h -- )  \ parse nibble to beat - MSB first for 4x4 matrix keypad entry
    5 1 do dup 1  4 i - lshift  and  4 i - rshift 
        [play-tone]   
(        if play-next-tone else next-tone then   )   \ works with tones-fixed
        // loop drop ;
: Nplayonce ( <nibbles> N -- <nibbles> N ) \ play sequence once from left to right (i.e. bottom of stack), eg  $F $E  2 Nplayonce
    >r r@ ndup nrev 0 do beat-nibble-msb loop r> ;
: Nplayloop ( N-bytes N -- N-bytes N ) \ loop sequence, plays left to right 
    visualize-beats reset-notecount reset-tonecount 
    begin Nplayonce ( readpot ) key? until key drop ;
: demo-uptown-funk 100 beatdur! 
    8 8 11 0   10 11 0 0
    8 8 11 0   11 15 8 0
    16 Nplayloop ;  \ 16 notes = 4 bars uptown funky bunch: https://www.youtube.com/watch?v=M1F0lBnsnkE


\ APPLICATION IDEA: Use a single 8-pin DIP to create 2 note combinations, and OLED spits out their binary encoding, 
\ But touchpad entry allows typing the beats in directly, and uC accepts them as the basis for looping.
\ THIS IS WHERE SOFTWARE BEATS HARDWARE!
\ With a mixed interface (software + hardware) we get the best of both worlds - optimized user interaction/interface, and precise timing independent of Microsoft Windows, and not requiring a computer to run.



\ TODO: Expand to a menu system:  ADD bar, EDIT, DELETE bar.  
\       Extract from keypad to fewest number of buttons
\       Add potentiometer for rate control
\       Incorporate keypad with 74165 shift-in monitor to allow controlling keypad with 3 input wires instead of 8
\       Incorporate OLED
\       Incorporate full menu system: Add bar, delete bar, fwd/back.   5 buttons   - < o > + 
\       Trigger keypad from interrupt to avoid need to continually monitor.
\       Run in code .avr, move to Attiny84 (11 pins) for dedicated keypad monitoring... but need Attiny comms to Nano

\ ======================================================================

: help-beats
  cr
  cr ." BEATS (help-beats) (help-inputbeats)"
  cr 
  cr ." hardware: requires Nano, either 74165 IC on nano pins 8,9,10, 8-bit dip or switches, buzzer on nano pin 4"
  cr ."         OR  Nano, 4x4 matrix keypad on nano pins 2-9, buzzer on nano pin 17"
  cr
  cr ." include   3iForth.fs  [ 3iShift.fs  3iKeypad.fs ] 3iSound-4x4.fs  3iBeats.fs   (requires Nano running 3iForth.avr)"
  cr ." 1 open-com   rx"
  cr
  cr ." init-beats  [ init-shiftin ]"
  cr ." init-percustones"
  cr ." metronome (variable beat with potentiometer)  [ 200 Nmetronome ] (quarter notes)  "
  cr ." demo-uptown-funk"
  cr
  cr ." N Nloopbeats ( hex digits, play from left to right)"
  cr ." beats-demo"
  cr
;

help-beats

\ V2 - 2021-05-26---Wed---07:52  extracted from Shift.fs.  Added keypad entry with software prompt for 2 bars
\ V3 - 2021-06-01---Tue---02:04  added: beat sequence with tones -- for two-tone drum for e.g. garbi -- bass and snare or clap & snap
\     Design: put a speaker with musical tones to simulate deep and high, e.g. bass and timpani drums etc.
\     Then the design cycles through executables of tone, matched to the positive beats, instead of just to pulse-hi
\     This is where ColorForth is better because it adds a few register elements to hold indexed elements/additional stack
\     In this case, e.g. there could be a stack with just these elements.  That's what memory is... 
\     the advantage in ColorForth is use of a registers instead of memory for this, i.e. it is much faster then
\ V4 - 2021-07-31---Sat---13:46  solidified into high color-depth percussion: ADDED ability to build tonal beats individually for 64-semiquavers (4 bar) phrases
